/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  Originally written by Art Vodolazsky <burivuh@crew.org.ru>
 *  Code clean-up and rewriting it to OOP method by 
 *  Alphex Kaänoken <kaanoken@crew.org.ru>
 *  after Andrew Crew Kuznetsov <andrew.crew.kuznetsov@gmail.com> rewrite it.
 *
 *  (c) Crew IT Research Labs 2004
 */

#include "../libmisc/types.h"
#include "../libmisc/error.h"
#include "../libmisc/utils.h"
#include "library.h"
#include "key_log_manager.h"

#include <X11/keysym.h>
#include <string.h>

int xklbuffer_free(struct _xkeylogbuffer *p);
int xklbuffer_free(struct _xkeylogbuffer *p)
{
  status err = SUCCESS;
  
  if (NULL != p) {
    if (NULL != p->buffer)
      p->buffer->uninit(p->buffer);
    else {
      err = WARNING;
    }
    free (p);
  }
  else
    err = WARNING;
  
  return err;
}

int xklbuffer_clear(struct _xkeylogbuffer *p);
int xklbuffer_clear(struct _xkeylogbuffer *p)
{
  int err;
	
  err=p->buffer->set_content(p->buffer, NULL);
  
  p->iterator = 0;
  
  return SUCCESS;
}

xkeylogbuffer *xkeylogbuffer_init (struct _xkeylogbuffer *p)
{
  p = (struct _xkeylogbuffer *) xnmalloc(sizeof(struct _xkeylogbuffer));
  if (NULL == p)
    return NULL;
  
  p->buffer = xstr_init (p->buffer, MAX_WORD_LENGTH);
  if (NULL == p->buffer) {
    free (p);
    return NULL;
  }
  p->iterator = 0;
  p->mode = KLB_INSERT_MODE;
  
  // Function mapping
  p->clear = xklbuffer_clear;
  p->free = xklbuffer_free;
  
  return p;
}

status xKeyLogManager_getBufferAction (xevent * event)
{
  status action;
  KeySym modifier = event->get_curr_modifier (event);
  KeySym keysym = event->get_curr_keysym (event);

  /*Cursor keys*/
  if (IsCursorKey (keysym))
    switch (keysym) {
    case XK_Right:
      action = KLB_MOVE_RIGHT;
      break;
    case XK_Left:
      action = KLB_MOVE_LEFT;
      break;
    default:
      action = KLB_CLEAR;
      break;
    }
  else
    /*KeyPad keys*/
    if (IsKeypadKey (keysym))
      switch (keysym) {
      case XK_KP_Right:
	action = KLB_MOVE_RIGHT;
	break;
      case XK_KP_Left:
	action = KLB_MOVE_LEFT;
	break;
      case XK_KP_Space:
	action = KLB_FLUSH;
	break;
      case XK_KP_Delete:
	action = KLB_DEL_SYM_RIGHT;
	break;
      case XK_KP_Insert:
	action = KLB_SWITCH_MODE;
	break;
      default:
	action = KLB_CLEAR;
	break;
      }
    else
      /*Func, Mod, PF, PrivateKeypad keys*/
      if (IsFunctionKey (keysym) || IsModifierKey (keysym) ||
	  IsPFKey (keysym) || IsPrivateKeypadKey (keysym))
	    action = KLB_NO_ACTION;
      else
	/*MiscFunc keys*/
	if (IsMiscFunctionKey (keysym))
	  if (keysym == XK_Insert) {
		if ((modifier == XK_Control_R)||(modifier == XK_Alt_R)||(modifier == XK_Shift_R))
	      action = KLB_CLEAR;
		else		  
	      action = KLB_SWITCH_MODE;
	  }
	  else
	    action = KLB_NO_ACTION;
	else
	  /*del, bkspace, tab, return, alpha & num keys*/
	  switch (keysym) {
	  case XK_Delete:
	    action = KLB_DEL_SYM_RIGHT;
	    break;
	  case XK_BackSpace:
	    action = KLB_DEL_SYM_LEFT;
	    break;
	  case XK_Tab:
	    action = KLB_CLEAR;
	    break;
	  case XK_Return:
	    action = KLB_CLEAR;
	    break;
	  case XK_Pause:
	  case XK_Escape:
	  case XK_Sys_Req:
	    action = KLB_NO_ACTION;
	    break;
  	  case XK_space:
		action = KLB_FLUSH;
	    break; 
	  case XK_exclam:
	  case XK_at:
	  case XK_numbersign:
	  case XK_ssharp:
	  case XK_dollar:
	  case XK_percent:
	  case XK_asciicircum:
	  case XK_ampersand:
	  case XK_asterisk:
	  case XK_parenleft:
	  case XK_parenright:
	  case XK_hyphen:
	  case XK_underscore:
	  case XK_equal:
	  case XK_plus:
	  case XK_minus:
	  case XK_slash:
	  case XK_bar:
	  case XK_backslash:
	  case XK_question:
	    action = KLB_FLUSH;
	    break;
	  default:	    
	    if ((modifier == XK_Control_R)||(modifier == XK_Alt_R))
	      action = KLB_NO_ACTION;
		else
		  action = KLB_ADD_SYM;	
	    break;
	  }

  return action;
}

status xKeyLogManager_getBindAction (xevent * event, xbindingtable * xbtable)
{
  status action;
  KeySym keysym = event->get_curr_keysym (event);
  KeySym modifier = event->get_curr_modifier (event);
  action = xbtable->find_bind_action (xbtable, keysym, modifier);
  if (action == -1)
    return BA_NONE;
  else
    return action;
}

status xKeyLogManager_flushBuffer (xkeylogbuffer * klb, xstr * xstring)
{
  char *str = (char *) xnmalloc (MAX_WORD_LENGTH);
  str = memset (str, NULLSYM, MAX_WORD_LENGTH);
  if (klb->iterator != 0) {
    str = memcpy (str, klb->buffer->content, (klb->iterator));
    xstring->set_content (xstring, str);
  }
  free (str);
  return SUCCESS;
}

char *xKeyLogManager_getStringFromEventQueue (xkeylogbuffer * klb,
					      xevent * evnt,
					      xifocus * ifocus,
					      xstr * xword,
					      xstr * xstring,
					      xswitchlang * xswlang, xbindingtable * xbtable, status lang)
{
  status action = KLB_NO_ACTION, err = SUCCESS;
  num itr;
  char sym;
  bool changed = FALSE;
  
  xword->set_content(xword, NULL);
  if (ifocus->has_chng(ifocus)) {
    klb->clear(klb);
    err = xstring->set_content(xstring, NULL);
  }

  while (evnt->chck_key_press(evnt)) {
    if (!evnt->is_native (evnt)) {
      changed = FALSE;
      continue;
    }
	
    action = xKeyLogManager_getBindAction (evnt, xbtable);

    xbtable->set_action(xbtable, action);
    if (action > 0) {	//have to perform some bind actions. Flush current buffer and return
      xKeyLogManager_flushBuffer (klb, xword);	  	
      return xstring->content;
    }
	
    action = xKeyLogManager_getBufferAction (evnt);
	
	//int state = (evnt->event.xkey.state & ShiftMask);
    switch (action) {
    case KLB_CLEAR:
      klb->clear (klb);
      xstring->set_content (xstring, NULL);
      send_error (DEBUG, "KLB_CLEAR\n", NULL);
      break;
    case KLB_FLUSH:
      sym = evnt->get_curr_char (evnt, xswlang);
      err = klb->buffer->add_ch_byindx (klb->buffer, sym, klb->iterator);
      klb->buffer->cat_content (klb->buffer, xstring);
      klb->clear (klb);
      send_error (DEBUG, "KLB_FLUSH\n", NULL);
	  if (strlen(xstring->content) > 0)
        send_error (DEBUG, "Current string=%s\n", &xstring->content);
      break;
    case KLB_SWITCH_MODE:
      if (KLB_INSERT_MODE == klb->mode)
        klb->mode = KLB_REPLACE_MODE;
      else
        klb->mode = KLB_INSERT_MODE;
      send_error (DEBUG, "KLB_SWITCH_MODE\n", NULL);
      break;
    case KLB_MOVE_LEFT:
      if (klb->iterator > 0)
        klb->iterator--;
      else {
        klb->clear (klb);
        xstring->set_content (xstring, NULL);
      }
      send_error (DEBUG, "KLB_MOVE_LEFT\n", NULL);
      break;
    case KLB_MOVE_RIGHT:
      if (klb->iterator < (MAX_WORD_LENGTH - 1))
        klb->iterator++;
      else {
        klb->clear (klb);
        xstring->set_content (xstring, NULL);
      }
      if (NULLSYM == klb->buffer->get_ch_byindx (klb->buffer, klb->iterator))
        klb->clear (klb);
      send_error (DEBUG, "KLB_MOVE_RIGHT\n", NULL);
      break;
    case KLB_DEL_SYM_LEFT:
      if (klb->iterator > 0) {
        for (itr = klb->iterator; itr <= strlen (klb->buffer->content); itr++)
          klb->buffer->content[itr - 1]=klb->buffer->content[itr];
        klb->iterator--;
      }
      else {
        klb->clear (klb);
        xstring->set_content (xstring, NULL);
      }
      send_error (DEBUG, "KLB_DEL_SYM_LEFT\n", NULL);
      break;
    case KLB_DEL_SYM_RIGHT:
      if (strlen (klb->buffer->content) < (MAX_WORD_LENGTH - 1))
        for (itr = klb->iterator; itr <= strlen (klb->buffer->content); itr++)
          klb->buffer->content[itr] = klb->buffer->content[itr + 1];
      else {
        klb->clear (klb);
        xstring->set_content (xstring, NULL);
      }
      send_error (DEBUG, "KLB_DEL_SYM_RIGHT\n", NULL);
      break;
    case KLB_ADD_SYM: 		
      sym = evnt->get_curr_char (evnt, xswlang);
      if (klb->mode == KLB_INSERT_MODE) {
        err = klb->buffer->add_ch_byindx (klb->buffer, sym, klb->iterator);
      }
      else {
        err = klb->buffer->set_ch_byindx (klb->buffer, sym, klb->iterator - 1);		
	  }
      if (err != SUCCESS) {
        send_error (ERROR, "Index is out of range!\n", NULL);
		klb->buffer->cat_content (klb->buffer, xstring);
		klb->clear (klb);
        send_error (DEBUG, "KLB_FLUSH\n", NULL);
		klb->buffer->add_ch_byindx (klb->buffer, sym, klb->iterator);

        break;
      }
      klb->iterator++;	
      xswlang->currlang = xswlang->get_curr_lang(xswlang, evnt->communicator->xDisplay);
      changed = TRUE;
      send_error (DEBUG, "KLB_ADD_SYM\n", NULL);
      break;
    default:
      send_error (DEBUG, "KLB_NO_ACTION\n", NULL);
      break;
    }	      	
  }
  
  if (changed) {
    xKeyLogManager_flushBuffer (klb, xword);
    return xword->content;
  }
  else
    return NULL;
}

status xKeyLogManager_bind (xbindingtable * xbtable, status key, status modifier, status action)
{
  status err = SUCCESS;
  KeySym key_ks = None, key_ks_alt = None, modifier_ks = None;
	
  switch (key) {
  case PAUSE_KEY:
    key_ks = XK_Pause;
    key_ks_alt = XK_Break;
    break;
  case SCROLL_LOCK_KEY:
    key_ks = XK_Scroll_Lock;
    key_ks_alt = None;
    break;
  case SYS_RQ_KEY:
    key_ks = XK_Print;
    key_ks_alt = XK_Sys_Req;
    break;
    break;
  case None:
    key_ks = None;
    key_ks_alt = None;
    break;
  default:
    send_error (WARNING, "Unknown key to bind\n", NULL);
    err = WARNING;
    break;
  }
  switch (modifier) {
  case SHIFT_KEY:
    modifier_ks = XK_Shift_R;
    break;
  case CONTROL_KEY:
    modifier_ks = XK_Control_R;
    break;
  case None:
    modifier_ks = None;
    break;
  default:
    send_error (WARNING, "Unknown modifier to bind\n", NULL);
    err = WARNING;
    break;
  }
  
  err = xbtable->bind_key(xbtable, key_ks, key_ks_alt, modifier_ks, action);
  
  return err;
}
