/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  Originally written by Art Vodolazsky <burivuh@crew.org.ru>
 *  rewritten by Alfeiks Kaänoken <akaanoken@gmail.com>
 *  after Andrew Crew Kuznetsov <andrew.crew.kuznetsov@gmail.com> rewrite it.
 *  (c) 2004, 2005, 2006
 */

/* IMPORTANT NOTES:*/
// 1)Work only with null-terminated strings

/*TODO: make -I flag when compiling, instead of the full path*/

#include "library.h"
#include "select_manager.h"
#include "../libmisc/types.h"
#include "../libmisc/error.h"
#include "../libmisc/utils.h"

#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>
#include <X11/Xdefs.h>
#include <X11/Xmu/Atoms.h>

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>

int handleXErrors (Display * d, XErrorEvent * e)
{
  return False;
}

/*-----xobj------*/

Display* xobj_get_display(struct _xobj *p);
Display* xobj_get_display(struct _xobj *p)
{
  return p->xDisplay;
}

int xobj_connect(xobj *p);
int xobj_connect(xobj *p)
{
  p->xDisplay=XOpenDisplay(NULL);
  if(!p->xDisplay){
    fprintf(stderr,"Cannot open display.\n");
    return ERROR;
  }
  return SUCCESS;
}

int xobj_set_err_handler(int (*handler) (Display *,XErrorEvent *));
int xobj_set_err_handler(int (*handler) (Display *,XErrorEvent *))
{
  XSetErrorHandler ((XErrorHandler) handler);
  return SUCCESS;
}

int xobj_create_win(struct _xobj *p);
int xobj_create_win(struct _xobj *p)
{
  p->xWindow=XCreateSimpleWindow(p->xDisplay,DefaultRootWindow(p->xDisplay),
				 0,0,1,1,0,0,0);
  if(!(p->xWindow)){
    fprintf(stderr,"Cannot create top level window.\n");
    return ERROR;
  }
  return SUCCESS;
}

Window xobj_get_win(struct _xobj *p);
Window xobj_get_win(struct _xobj *p)
{
  bool is_destroyed;
  xevent *xevnt = NULL;
  Window win;
  
  xevnt = xevent_init (xevnt);
  xevnt->set_communicator(xevnt, p);
	
  is_destroyed = xevnt->chck_destroy_notify(xevnt);
  if (is_destroyed) {
    fprintf(stderr,"Processed window is destroyed already.\n");
    win=ERROR;
  }
  else
    win=p->xWindow;

  xevnt->uninit(xevnt);
  return win;
}

Window xobj_get_app_root_win(struct _xobj *p);
Window xobj_get_app_root_win(struct _xobj *p)
{
  Window backup_window;
  Window root_window;
  Window parent_window;
  Window window=p->get_win(p);
  Window *children_window_list;
  unsigned int nchildren;
  bool exit_flag=FALSE;
	
  if (window==ERROR)
    return ERROR;
  
  backup_window = window;
  
  while(!exit_flag) {
    if(!XQueryTree
       (p->get_display(p), window, &root_window,
	&parent_window, &children_window_list, &nchildren))
      return 0;

    if(children_window_list)
      XFree(children_window_list);

    if (parent_window == root_window)
      exit_flag = TRUE;
    else{
      backup_window = window;
      window = parent_window;
    }
  }

  return backup_window;
}

Window xobj_find_win_byatom(struct _xobj *p,char *atom);
Window xobj_find_win_byatom(struct _xobj *p,char *atom)
{
  Atom atom_;
  Atom type=None;
  int format;
  unsigned long nitems,after;
  unsigned char *data=NULL;
  Window app_root_window;
  Window tmp_window=0,win;

  win=p->get_win(p);
  if (win==ERROR)
    return ERROR;

  if(win==RootWindow(p->get_display(p),DefaultScreen(p->get_display(p))))
    return win;

  atom_=XInternAtom(p->xDisplay,atom,True);
  if (!atom)
    return p->get_win(p);

  app_root_window=p->get_app_root_win(p);
  if(app_root_window==ERROR)
    return ERROR;
  else if(app_root_window==0)
    return 0;

  XGetWindowProperty(p->get_display(p),app_root_window,atom_,0,0,False,
		     AnyPropertyType,&type,&format,&nitems,&after,&data);

  if(data)
    XFree (data);

  if(type)
    return app_root_window;

  tmp_window=p->find_child_byatom(p->get_display(p),&app_root_window,atom_);

  if (tmp_window==ERROR)
    return ERROR;
  else if(!tmp_window)
    tmp_window=p->get_win(p);

  return tmp_window;
}

Window xobj_find_child_byatom(Display *xdsp,Window *xwin,Atom atom);
Window xobj_find_child_byatom(Display *xdsp,Window *xwin,Atom atom)
{
  Atom type=None;
  int format;
  unsigned long nitems, after;
  unsigned char *data = NULL;
  Window tmp_window = 0, root_window, parent_window;
  Window *children_window_list;
  unsigned int nchildren;
  unsigned int i;

  if (xwin<=0)
    return ERROR;

  if(!XQueryTree(xdsp,*xwin,&root_window,&parent_window,
		  &children_window_list,&nchildren))
    return 0;

  for (i = 0;!tmp_window && (i<nchildren);i++){
    data = NULL;
    XGetWindowProperty(xdsp,children_window_list[i],atom,
			0,0,False,AnyPropertyType,&type,
			&format,&nitems,&after,&data);
    if (data)
      XFree (data);
    if (type)
      tmp_window=children_window_list[i];
  }

  for(i=0;!tmp_window && (i<nchildren);i++)
    tmp_window=xobj_find_child_byatom(xdsp,&children_window_list[i],atom);

  if(children_window_list)
    XFree(children_window_list);

  return tmp_window;
}

XClassHint *xobj_get_wm_class(struct _xobj *p);
XClassHint *xobj_get_wm_class(struct _xobj *p)
{
  XClassHint *wm_class=NULL;
  Window window;

  window=p->find_win_byatom(p,"WM_CLASS");
  if (window==ERROR)
    return NULL;
  if (!window)
    window=p->get_win(p);

  wm_class=XAllocClassHint();
  if (!XGetClassHint(p->get_display(p),window,wm_class)){
    fprintf(stderr,"Cannot get WM_CLASS\n");
    XFree(wm_class->res_class);
    XFree(wm_class->res_name);
    XFree(wm_class);
	usleep(500000); // My optimal delay for this Error.
    return NULL;
  }

  return wm_class;
}

char *xobj_get_wmclass_class(struct _xobj *p);
char *xobj_get_wmclass_class(struct _xobj *p)
{
  char *string=NULL;
  XClassHint *wm_class;

  wm_class=p->get_wm_class(p);
  if (!wm_class)
    return NULL;

  string=xnmalloc(strlen(wm_class->res_class)+1);
  string=memset(string,'\0',strlen(wm_class->res_class)+1);
  string=memcpy(string,wm_class->res_class,strlen(wm_class->res_class));

  XFree(wm_class->res_class);
  XFree(wm_class->res_name);
  XFree(wm_class);

  return string;
}

int xobj_set_event_mask(struct _xobj *p,mask event_mask);
int xobj_set_event_mask(struct _xobj *p,mask event_mask)
{  
  Window win=p->get_win(p);  ;
  if (win!=ERROR)
    XSelectInput(p->get_display(p),win,event_mask);
  else
    return ERROR;

  return SUCCESS;
}

int xobj_uninit(struct _xobj *p);
int xobj_uninit(struct _xobj *p)
{
  if(p)
    free(p);
  else
    return WARNING;
  return SUCCESS;
}

xobj *xobj_init (xobj *p)
{
  p=NULL;
  p=(xobj *) xnmalloc(sizeof(struct _xobj));
  p->xDisplay=NULL;
  p->xWindow=WRONG_NUM;
  p->xScreenNum=WRONG_NUM;

  /*functions mapping*/
  p->get_display=xobj_get_display;
  p->connect=xobj_connect;
  p->set_err_handler=xobj_set_err_handler;
  p->create_win=xobj_create_win;
  p->get_win=xobj_get_win;
  p->get_app_root_win=xobj_get_app_root_win;
  p->find_win_byatom=xobj_find_win_byatom;
  p->find_child_byatom=xobj_find_child_byatom;
  p->get_wm_class=xobj_get_wm_class;
  p->get_wmclass_name=xobj_get_wmclass_class;/*don't forget to free after use*/
  p->get_wmclass_class=xobj_get_wmclass_class;/*don't forget to free after use*/
  p->set_event_mask=xobj_set_event_mask;
  p->uninit=xobj_uninit;

  return p;
}

/*------xstr-------*/ 

int xstr_set_key_code(struct _xstr *p, int lang, xswitchlang *swlng);
int xstr_set_key_code(struct _xstr *p, int lang, xswitchlang *swlng)
{
  unsigned int charIndex;
  mask keycode_mod;
  mask shift_mod;
  KeySym ks;
  char element[2]={'\0','\0'};
  xevent* xevnt = NULL;
  
  if (!(p->home->get_display(p->home))) 
    return INCORRECT_OPERATION;
  if ((WRONG_SIZE==(p->currentSize)) || (0==(p->currentSize)))
    return INCORRECT_OPERATION;

  p->keycodeModifiers=memset(p->keycodeModifiers,0,sizeof (mask) * (p->currentSize));

  xevnt = xevent_init(xevnt);
  for(charIndex=0;charIndex < strlen(p->content);charIndex++){
    element[0]=p->content[charIndex];

    ks=xevnt->char_to_keysym(element[0]);
	if (lang == RUSSIAN) {   
	  switch (swlng->rus_win_keyboard_group) {
		case 0:
		  keycode_mod = Group0Mask;
		  break;
		case 1:
		  keycode_mod = Group1Mask;
		  break;
		case 2:
		  keycode_mod = Group2Mask;
		  break;
		case 3:
		  keycode_mod = Group3Mask;
		  break;
	  }
	}
	else if (lang == ENGLISH) {  
	  switch (swlng->usa_eng_keyboard_group) {
		case 0:
		  keycode_mod = Group0Mask;
		  break;
		case 1:
		  keycode_mod = Group1Mask;
		  break;
		case 2:
		  keycode_mod = Group2Mask;
		  break;
		case 3:
		  keycode_mod = Group3Mask;
		  break;
	  }
	}	
	if (isUpperNonAlphaCyr (element[0]) | isupper(element[0]))
	  shift_mod = ShiftMask;
	else
	  shift_mod = None;

	p->keycodeModifiers[charIndex] = (keycode_mod | shift_mod);
    p->keycode[charIndex]=XKeysymToKeycode(p->home->get_display(p->home),ks);
	// Xlib (error '<') patch >>
    if ((ks == XK_less) && (p->keycode[charIndex] == 94))
	  p->keycode[charIndex] = 59;	
	// <<
  }
  	
  xevnt->uninit(xevnt);	 
  return SUCCESS;
}

// This Function Not Use!
int xstr_to_ascii(struct _xstr *p);
int xstr_to_ascii(struct _xstr *p)
{
  char *array=p->content;
  size arraySize=WRONG_SIZE;
  num counter=WRONG_NUM,index = WRONG_NUM;
  char element[CYR_BYTES_PER_SYM+1]={'\0','\0','\0'};
  
  if(!array)
    return 0;

  arraySize=strlen(array);
  counter=0;
  index=0;
  
  if(isascii(*(array))){
    //p->set_key_code(p,None);
    return ENGLISH;
  }

  while(index<=(arraySize)/CYR_BYTES_PER_SYM){
    element[0]=array[counter];
    element[1]=array[counter+1];
    
    array[index]=get_ascii(element);
    counter+=CYR_BYTES_PER_SYM;
    index++;
  }
  memset((array+index),'\0',p->currentSize-index);
  //p->set_key_code(p,Group1Mask);
  
  return RUSSIAN;
}

int xstr_set_content(struct _xstr *p,char *src);
int xstr_set_content(struct _xstr *p,char *src)
{
  size_t sizeToChange;
  size_t sizeOld;
  size_t fullsizeOld;
  sizeOld=strlen(p->content);
  fullsizeOld=p->currentSize;
  if(!src){
    memset(p->content,'\0',fullsizeOld);
    return 0;
  }

  sizeToChange=strlen(src);
  
  if(sizeOld==WRONG_SIZE)
    sizeOld=0;
  
  if(!sizeToChange){
    if(sizeOld>0){
      memset(p->content,'\0',sizeOld);
      memset(p->keycode,0,sizeof(KeyCode) * sizeOld);
      memset(p->keycodeModifiers,0,sizeof(mask) * sizeOld);
      return SUCCESS;
    }
    else
      return WARNING;
  }
  else{
    if(sizeToChange>=fullsizeOld){
      p->content=xncalloc((int) p->content,sizeToChange+1);
      p->keycode=xncalloc((int) p->keycode,sizeof(KeyCode) * (sizeToChange+1));
      p->keycodeModifiers=xncalloc((int) p->keycodeModifiers,
				    sizeof(mask) * (sizeToChange+1));
      memset(p->content,'\0',sizeToChange+1);
      memset(p->keycode,0,sizeof(KeyCode) * (sizeToChange+1));
      memset(p->keycodeModifiers,0,sizeof(mask) * (sizeToChange+1));
      p->currentSize = sizeToChange + 1;
    }
    else{
      memset(p->content,'\0',fullsizeOld);
      memset(p->keycode,0,sizeof(KeyCode) * fullsizeOld);
      memset(p->keycodeModifiers,0,sizeof(mask) * fullsizeOld);
    }
    memcpy(p->content,src,sizeToChange);
  }

  return SUCCESS;
}

int xstr_cat_content(struct _xstr *p,struct _xstr *dest);
int xstr_cat_content(struct _xstr *p,struct _xstr *dest)
{
  size_t len_src=p->get_len(p);
  size_t len_dest=dest->get_len(dest);
  char *tmp;

  if (!len_src)
    return 0;
  tmp=xnmalloc(len_dest+len_src+1);
  
  memset(tmp,'\0',len_dest+len_src+1);
  strncat(tmp,dest->content,len_dest);
  strncat(tmp,p->content,len_src);
  dest->set_content(dest,tmp);
  
  free(tmp);
  return SUCCESS;  
}

int xstr_add_ch_byindx(struct _xstr *p,char sym,int index);
int xstr_add_ch_byindx(struct _xstr *p,char sym,int index)
{
  char *buffer=NULL;
  int ai=0;
  
  if(index >= p->currentSize - 2)
    return ERROR;

  if(index==(p->currentSize-2)){
    p->currentSize+=p->currentSize;
    p->content=xncalloc((int) p->content,sizeof(char)*(p->currentSize));
    p->keycode=xncalloc((int) p->keycode,sizeof(KeyCode) * (p->currentSize));
    p->keycodeModifiers=xncalloc((int) p->keycodeModifiers,
					  sizeof (mask) * (p->currentSize));
  }

  ai=strlen(p->content);
  if(ai>index){
    buffer=(char *)xnmalloc(p->currentSize);
    buffer=memcpy(buffer,p->content+index,ai-index);
  }
  else
    buffer = NULL;
  
  p->content[index]=sym;
  p->content[index+1]='\0';
  if(buffer){
    p->content=strcat(p->content,buffer);
    free(buffer);
  }
  
  return SUCCESS;
}

int xstr_set_ch_byindx(struct _xstr *p,char sym,int index);
int xstr_set_ch_byindx(struct _xstr *p,char sym,int index)
{
  if (index >= p->currentSize - 2)
    return ERROR;

  p->keycode=memset(p->keycode,'\0',p->currentSize);
  p->keycodeModifiers=memset(p->keycodeModifiers,'\0',p->currentSize);
  p->content[index]=sym;
  
  return SUCCESS;
}

char xstr_get_ch_byindx(struct _xstr *p,int index);
char xstr_get_ch_byindx(struct _xstr *p,int index)
{
  if(index<p->currentSize)
    return p->content[index];
  else
    return '\0';
}

size_t xstr_get_len(struct _xstr *p);
size_t xstr_get_len(struct _xstr *p)
{
  if(p->content)
    return strlen(p->content);
  else 
    return 0;
}

int xstr_uninit(struct _xstr *p);
int xstr_uninit(struct _xstr *p)
{
  int err=SUCCESS;

  if(p){
    if((p->home->uninit(p->home))!=SUCCESS)
      err=WARNING;
    if ((p->keycodeModifiers))
      free((p->keycodeModifiers));
    else
      err=WARNING;
    if(p->keycode)
      free(p->keycode);
    else
      err=WARNING;
    if(p->content)
      free(p->content);
    else
      err=WARNING;
    free (p);
  }
  else
    return WARNING;

  return err;
}

xstr *xstr_init (xstr *p,size init_size)
{
  p=NULL;
  p=(xstr *)xnmalloc (sizeof(struct _xstr));
  p->content=NULL;

  if (!init_size)
    p->currentSize=WRONG_SIZE;
  else
    p->currentSize=init_size;

  if(init_size){
    p->content=(char *)xnmalloc(init_size);
    p->keycode=(KeyCode *)xnmalloc(sizeof(KeyCode) * init_size);
    p->keycodeModifiers=(mask *)xnmalloc(sizeof(mask) * init_size);
    p->home=xobj_init(p->home);
    memset(p->content,'\0',init_size);
    memset(p->keycode,0,init_size);
    memset(p->keycodeModifiers,0,init_size);
  }

  /*functions mapping*/
  p->set_key_code=xstr_set_key_code;
  p->to_ascii=xstr_to_ascii;
  p->set_content=xstr_set_content;
  p->cat_content=xstr_cat_content;
  p->add_ch_byindx=xstr_add_ch_byindx;
  p->set_ch_byindx=xstr_set_ch_byindx;
  p->get_ch_byindx=xstr_get_ch_byindx;
  p->get_len=xstr_get_len;
  p->uninit=xstr_uninit;

  return p;
}

/*--------xifocus--------*/

int xifocus_uninit(struct _xifocus *p);
int xifocus_uninit(struct _xifocus *p)
{
  if(p){
    p->owner->uninit(p->owner)!=SUCCESS;
    while(p->ptr_num>0)
      p->ptr[--(p->ptr_num)]->uninit(p->ptr[--(p->ptr_num)]);
    if(p->ptr)
      free(p->ptr);
    free(p);
  }
  else
    return WARNING;
  
  return SUCCESS;
}

int xifocus_update(struct _xifocus *p);
int xifocus_update(struct _xifocus *p)
{
  Window oldwin=p->owner->get_win(p->owner);
  bool isSameScreen;
  int dummy;
  unsigned int dummyU;
  Window rootW=0,childW=0;
  char *new_app_name, *old_app_name;

  old_app_name = p->owner->get_wmclass_name(p->owner);
	
  if(!oldwin)
    return INCORRECT_OPERATION;

  XGetInputFocus(p->owner->xDisplay, &(p->owner->xWindow), &(p->cur_status));
  
  if(None==p->owner->xWindow) {
    p->owner->xWindow=oldwin;
    return ERROR;
  }	  
  if(oldwin!=p->owner->get_win(p->owner)){
    p->changed=TRUE;  
	new_app_name = p->owner->get_wmclass_name(p->owner);
	if ((!new_app_name)||(!old_app_name)||(strcmp(new_app_name, old_app_name))){	  
      send_error(DEBUG,"New Processed Window With Name '%s'.\n",&new_app_name);      
	}
	if (new_app_name!=NULL) free(new_app_name);
  }
  else {
    p->changed=FALSE;	
  }  
  if (old_app_name!=NULL) free(old_app_name);
  
  while((p->ptr_num)>0){
    p->ptr[--(p->ptr_num)]->uninit(p->ptr[(p->ptr_num)]);
  }	  
  if((p->ptr)!=NULL){
    free(p->ptr);
    p->ptr=NULL;
  }
  
  childW=p->owner->get_win(p->owner);
  if (childW==ERROR)
    return ERROR;

  if((p->cur_status==RevertToPointerRoot)){
    if(p->owner->get_win(p->owner)==ERROR)
      return ERROR;
    
    isSameScreen=XQueryPointer(p->owner->get_display(p->owner),childW,&rootW,
			       &childW,&dummy,&dummy,&(dummy),&(dummy),&dummyU);
    if (isSameScreen==TRUE){
      while(childW!=0){

		(p->ptr_num)++;
		p->ptr=realloc(p->ptr,sizeof(void*)*(p->ptr_num));
		if (p->ptr == NULL) {
		  return ERROR;
		}
		p->ptr[(p->ptr_num)-1]=xobj_init(p->ptr[(p->ptr_num)-1]);
	
		p->ptr[(p->ptr_num)-1]->xDisplay=p->owner->get_display(p->owner);
		p->ptr[(p->ptr_num)-1]->xWindow=childW;
		
		if((p->ptr[(p->ptr_num) - 1]->get_win(p->ptr[(p->ptr_num) - 1]))==ERROR)
          return ERROR;
		isSameScreen=XQueryPointer(p->owner->get_display(p->owner),childW,&rootW,
				   &childW,&dummy,&dummy,&(dummy),&(dummy),&dummyU);
      }
    }
  }
  return SUCCESS;
}

bool xifocus_has_chng(struct _xifocus *p);
bool xifocus_has_chng(struct _xifocus *p)
{
  return p->changed==TRUE;
}

xifocus *xifocus_init(xifocus *p);
xifocus *xifocus_init(xifocus *p)
{
  p=NULL;

  p=(xifocus *)xnmalloc(sizeof(struct _xifocus));
  p->owner=xobj_init(p->owner);
  p->ptr=NULL;
  p->ptr_num=0;
  p->cur_status=WRONG_NUM;
  p->changed=FALSE;

  /*functions mapping*/
  
  p->has_chng=xifocus_has_chng;
  p->update=xifocus_update;
  p->uninit=xifocus_uninit;
  
  return p;
}

/*--------xevent--------*/

int xevent_set_communicator(struct _xevent *p,xobj *set);
int xevent_set_communicator(struct _xevent *p,xobj *set)
{
  p->communicator = memcpy (p->communicator, set, sizeof (xobj));
  return SUCCESS;
}

Window xevent_get_communicator_win(struct _xevent *p);
Window xevent_get_communicator_win(struct _xevent *p)
{
  return xobj_get_win (p->communicator);
}

Window xevent_get_event_win(struct _xevent *p);
Window xevent_get_event_win(struct _xevent *p)
{
  return p->event.xkey.window;
}

int xevent_update_by_focus(struct _xevent *p,xifocus *f);
int xevent_update_by_focus(struct _xevent *p,xifocus *f)
{
  return 0; // CHECK FUNCTION CODE !!!
}

int xevent_send_xkey(struct _xevent *p,KeyCode kc,mask modifiers);
int xevent_send_xkey(struct _xevent *p,KeyCode kc,mask modifiers)
{
  status err = SUCCESS;
  Window win;
  Display *disp;

  if (!(p->communicator->get_display (p->communicator)))
    return INCORRECT_OPERATION;
 
  p->event.type = KeyPress;
  p->event.xkey.type = KeyPress;
  p->event.xkey.window = p->communicator->xWindow;
  p->event.xkey.root =
    RootWindow (p->communicator->get_display (p->communicator),
		DefaultScreen (p->communicator->get_display (p->communicator)));
  p->event.xkey.subwindow = None;
  p->event.xkey.same_screen = True;
  p->event.xkey.display = p->communicator->xDisplay;
  p->event.xkey.state = modifiers;
  p->event.xkey.keycode = kc;
  p->event.xkey.time = CurrentTime;

  win = p->communicator->get_win (p->communicator);
  if (win == ERROR)
    return ERROR;
  
  err = XSendEvent (p->communicator->get_display (p->communicator), win,
		    TRUE, CLIENT_EVENT_MASK, &(p->event));
  if (err) {
    p->event.type = KeyRelease;
    p->event.xkey.type = KeyRelease;
    p->event.xkey.time = CurrentTime;

    win = xobj_get_win (p->communicator);
    if (win == ERROR)
      return ERROR;
	disp=p->communicator->get_display (p->communicator);
	if (disp == NULL)
      return ERROR;
    err = XSendEvent (p->communicator->get_display (p->communicator),
		      win, TRUE, KeyReleaseMask, &(p->event));
    if (!err)
      return INCORRECT_OPERATION;
  }
  else
    return INCORRECT_OPERATION;

  return err;
}

int xevent_send_event(struct _xevent *p,XEvent *event,mask emask);
int xevent_send_event(struct _xevent *p,XEvent *event,mask emask)
{
  status err;

  err = XSendEvent (p->communicator->get_display (p->communicator),
		    xobj_get_win (p->communicator), TRUE, emask, event);
  if (!err)
    return INCORRECT_OPERATION;
  else
    return SUCCESS;
}

int xevent_send_string(struct _xevent *p,xstr *sp);
int xevent_send_string(struct _xevent *p,xstr *sp)
{
  unsigned int charIndex = 0;
  status err = SUCCESS;

  for (charIndex = 0; charIndex < strlen (sp->content); charIndex++) {
    err = xevent_send_xkey (p, sp->keycode[charIndex],
			   sp->keycodeModifiers[charIndex]);
    if (err == ERROR)
      return ERROR;
  }
  XFlush (p->communicator->xDisplay);

  return err;
}

int xevent_put_back_to_queue(struct _xevent *p, XEvent *event);
int xevent_put_back_to_queue(struct _xevent *p, XEvent *event)
{
  status err;

  err = XPutBackEvent (p->communicator->get_display (p->communicator), event);
  if (!err)
    return INCORRECT_OPERATION;
  else
    return SUCCESS;
}

bool xevent_chck_key_release(struct _xevent *p);
bool xevent_chck_key_release(struct _xevent *p)
{
  status err;

  err = XCheckMaskEvent (p->communicator->xDisplay,
			 KeyReleaseMask, &p->event);
  if (err)
    return TRUE;
  else
    return FALSE;
}

bool xevent_chck_key_press(struct _xevent *p);
bool xevent_chck_key_press(struct _xevent *p)
{
  status err;

  err = XCheckMaskEvent (p->communicator->get_display (p->communicator),
			 KeyPressMask, &p->event);
  if (err)
    return TRUE;
  else
    return FALSE;
}

bool xevent_chck_destroy_notify(struct _xevent *p);
bool xevent_chck_destroy_notify(struct _xevent *p)
{
  status err;
  err = XCheckMaskEvent (p->communicator->get_display(p->communicator),
			 StructureNotifyMask, &p->event);
  if (err && (p->event.type == DestroyNotify))
    return TRUE;
  else
    return FALSE;
}

bool xevent_is_native(struct _xevent *p);
bool xevent_is_native(struct _xevent *p)
{
  if (p->event.xkey.send_event == TRUE)
    return FALSE;
  else
    return TRUE;
}

char xevent_get_curr_char(struct _xevent *p, struct _xswitchlang *xlang);
char xevent_get_curr_char(struct _xevent *p, struct _xswitchlang *xlang)
{

  KeySym ks;
  int nbytes;
  char str[MAX_WORD_LENGTH];
  str[0] = NULLSYM;
  char *strPtr = NULL;
  int state = (p->event.xkey.state & ShiftMask);
  ks = XKeycodeToKeysym(p->communicator->xDisplay, p->event.xkey.keycode, state);
  str[0] = p->keysym_to_char (ks);
  return str[0];  
}

KeySym xevent_get_curr_keysym(struct _xevent *p);
KeySym xevent_get_curr_keysym(struct _xevent *p)
{
  KeySym ks;
  int nbytes;
  char *str = (char *) xnmalloc (MAX_WORD_LENGTH);
  nbytes = XLookupString (&(p->event.xkey), str, MAX_WORD_LENGTH, &ks, NULL);
  free (str);
	
  return ks;
}

KeySym xevent_get_curr_modifier(struct _xevent *p);
KeySym xevent_get_curr_modifier(struct _xevent *p)
{
  KeySym ks;

  if ((p->event.xkey.state) & ControlMask)
    ks = XK_Control_R;
  else if ((p->event.xkey.state) & ShiftMask)
    ks = XK_Shift_R;
  else if ((p->event.xkey.state) & Mod1Mask)
    ks = XK_Alt_R;
  else
    ks = None;
  return ks;
}

// Function not use!!!
KeySym xevent_RuWinlayout_ks_to_ENlayout_ks(KeySym ks);
KeySym xevent_RuWinlayout_ks_to_ENlayout_ks(KeySym ks)
{
  KeySym ks_return = ks;
  num iterator;
  KeySym ruswin[] = { 
	//Shift Numeric
	XK_Cyrillic_IO, XK_exclam, XK_quotedbl, XK_numerosign,  
	XK_percent, XK_question, XK_asterisk, XK_parenleft, 
	XK_parenright, XK_underscore, XK_plus, XK_slash,
	//Numeric  
	XK_Cyrillic_io, XK_1, XK_2, XK_3, XK_4, 
	XK_5, XK_6, XK_7, XK_8, XK_9, 
	XK_0, XK_minus, XK_equal, XK_backslash,
    //Shift Character 	  
	XK_Cyrillic_SHORTI, XK_Cyrillic_TSE, XK_Cyrillic_U, XK_Cyrillic_KA, 
	XK_Cyrillic_IE, XK_Cyrillic_EN, XK_Cyrillic_GHE, XK_Cyrillic_SHA, 
	XK_Cyrillic_SHCHA, XK_Cyrillic_ZE, XK_Cyrillic_HA, XK_Cyrillic_HARDSIGN,
	XK_Cyrillic_EF, XK_Cyrillic_YERU, XK_Cyrillic_VE, XK_Cyrillic_A,
	XK_Cyrillic_PE, XK_Cyrillic_ER, XK_Cyrillic_O, XK_Cyrillic_EL,
    XK_Cyrillic_DE, XK_Cyrillic_ZHE, XK_Cyrillic_E, XK_Cyrillic_YA,
	XK_Cyrillic_CHE, XK_Cyrillic_ES, XK_Cyrillic_EM, XK_Cyrillic_I,
	XK_Cyrillic_TE, XK_Cyrillic_SOFTSIGN, XK_Cyrillic_BE, XK_Cyrillic_YU, 
	XK_comma,	  
	//Character
    XK_Cyrillic_shorti, XK_Cyrillic_tse, XK_Cyrillic_u, XK_Cyrillic_ka, 
	XK_Cyrillic_ie, XK_Cyrillic_en, XK_Cyrillic_ghe, XK_Cyrillic_sha, 
	XK_Cyrillic_shcha, XK_Cyrillic_ze, XK_Cyrillic_ha, XK_Cyrillic_hardsign,
	XK_Cyrillic_ef, XK_Cyrillic_yeru, XK_Cyrillic_ve, XK_Cyrillic_a,
	XK_Cyrillic_pe, XK_Cyrillic_er, XK_Cyrillic_o, XK_Cyrillic_el,
    XK_Cyrillic_de, XK_Cyrillic_zhe, XK_Cyrillic_e, XK_Cyrillic_ya,
	XK_Cyrillic_che, XK_Cyrillic_es, XK_Cyrillic_em, XK_Cyrillic_i,
	XK_Cyrillic_te, XK_Cyrillic_softsign, XK_Cyrillic_be, XK_Cyrillic_yu, 
	XK_period	  	  
  };
  
  KeySym eng[] = { 
	//Shift Numeric 14
	XK_asciitilde, XK_exclam, XK_at, XK_numbersign, 
	XK_percent, XK_ampersand, XK_asterisk, XK_parenleft, 
	XK_parenright, XK_underscore, XK_plus, XK_slash,
	//Numeric 14  
	XK_grave, XK_1, XK_2, XK_3, XK_4, 
	XK_5, XK_6, XK_7, XK_8, XK_9, 
	XK_0, XK_minus, XK_equal, XK_backslash,  
    //Shift Character 33 	  
	XK_Q, XK_W, XK_E, XK_R, 
	XK_T, XK_Y, XK_U, XK_I, 
	XK_O, XK_P, XK_braceleft, XK_braceright,
	XK_A, XK_S, XK_D, XK_F,
	XK_G, XK_H, XK_J, XK_K,
    XK_L, XK_colon, XK_quotedbl, XK_Z,
	XK_X, XK_C, XK_V, XK_B,
	XK_N, XK_M, XK_less, XK_greater, 
	XK_question,	  
	//Character 33
    XK_q, XK_w, XK_e, XK_r, 
	XK_t, XK_y, XK_u, XK_i, 
	XK_o, XK_p, XK_bracketleft, XK_bracketright,
	XK_a, XK_s, XK_d, XK_f,
	XK_g, XK_h, XK_j, XK_k,
    XK_l, XK_semicolon, XK_apostrophe, XK_z,
	XK_x, XK_c, XK_v, XK_b,
	XK_n, XK_m, XK_comma, XK_period, 
	XK_slash	  	  
  };
  
  for (iterator = 0; iterator < 94; iterator++)
    if (ks == ruswin[iterator]) {
	  ks_return = eng[iterator];
	  break;
    }	
  return ks_return;
}

char xevent_keysym_to_char(KeySym ks);
char xevent_keysym_to_char(KeySym ks)
{
  num iterator;
  char res = NULLSYM;
  KeySym ksym[] = {   
	//Shift Numeric 14
	XK_asciitilde, XK_exclam, XK_at, XK_numbersign, XK_dollar, 
	XK_percent, XK_asciicircum, XK_ampersand, XK_asterisk, XK_parenleft, 
	XK_parenright, XK_underscore, XK_plus, XK_bar,
	//Numeric 14  
	XK_grave, XK_1, XK_2, XK_3, XK_4, 
	XK_5, XK_6, XK_7, XK_8, XK_9, 
	XK_0, XK_minus, XK_equal, XK_backslash,  
    //Shift Character 33 	  
	XK_Q, XK_W, XK_E, XK_R, 
	XK_T, XK_Y, XK_U, XK_I, 
	XK_O, XK_P, XK_braceleft, XK_braceright,
	XK_A, XK_S, XK_D, XK_F,
	XK_G, XK_H, XK_J, XK_K,
    XK_L, XK_colon, XK_quotedbl, XK_Z,
	XK_X, XK_C, XK_V, XK_B,
	XK_N, XK_M, XK_less, XK_greater, 
	XK_question,	  
	//Character 33
    XK_q, XK_w, XK_e, XK_r, 
	XK_t, XK_y, XK_u, XK_i, 
	XK_o, XK_p, XK_bracketleft, XK_bracketright,
	XK_a, XK_s, XK_d, XK_f,
	XK_g, XK_h, XK_j, XK_k,
    XK_l, XK_semicolon, XK_apostrophe, XK_z,
	XK_x, XK_c, XK_v, XK_b,
	XK_n, XK_m, XK_comma, XK_period, 
	XK_slash,
    // Space 1
    XK_space	
  };	
  char ch[] = { 

      '~', '!', '@', '#', '$',
	  '%', '^', '&', '*', '(',
	  ')', '_', '+', '|', 
	  
	  '`', '1', '2', '3', '4',
	  '5', '6', '7', '8', '9',
	  '0', '-', '=', '\\',
	  
	  'Q', 'W', 'E', 'R',
	  'T', 'Y', 'U', 'I',
	  'O', 'P', '{', '}',
	  'A', 'S', 'D', 'F',
	  'G', 'H', 'J', 'K',
	  'L', ':', '\"', 'Z',
	  'X', 'C', 'V', 'B',
	  'N', 'M', '<', '>',
	  '?',
	  
	  'q', 'w', 'e', 'r',
	  't', 'y', 'u', 'i',
	  'o', 'p', '[', ']',
	  'a', 's', 'd', 'f',
	  'g', 'h', 'j', 'k',
	  'l', ';', '\'', 'z',
	  'x', 'c', 'v', 'b',
	  'n', 'm', ',', '.',
	  '/', 
	  
	  ' '
  };

  for (iterator = 0; iterator < 95; iterator++)
    if (ks == ksym[iterator]) {
      res = ch[iterator];
      break;
    }
  return res;
}


KeySym xevent_char_to_keysym(char ch);
KeySym xevent_char_to_keysym(char ch)
{
  num iterator;
  KeySym res = None; 
  KeySym eng[] = { 
	//Shift Numeric 14
	XK_asciitilde, XK_exclam, XK_at, XK_numbersign, XK_dollar,
	XK_percent, XK_asciicircum, XK_ampersand, XK_asterisk, XK_parenleft, 
	XK_parenright, XK_underscore, XK_plus, XK_bar,
	//Numeric 14  
	XK_grave, XK_1, XK_2, XK_3, XK_4, 
	XK_5, XK_6, XK_7, XK_8, XK_9, 
	XK_0, XK_minus, XK_equal, XK_backslash,  
    //Shift Character 33 	  
	XK_Q, XK_W, XK_E, XK_R, 
	XK_T, XK_Y, XK_U, XK_I, 
	XK_O, XK_P, XK_braceleft, XK_braceright,
	XK_A, XK_S, XK_D, XK_F,
	XK_G, XK_H, XK_J, XK_K,
    XK_L, XK_colon, XK_quotedbl, XK_Z,
	XK_X, XK_C, XK_V, XK_B,
	XK_N, XK_M, XK_less, XK_greater, 
	XK_question,	  
	//Character 33
    XK_q, XK_w, XK_e, XK_r, 
	XK_t, XK_y, XK_u, XK_i, 
	XK_o, XK_p, XK_bracketleft, XK_bracketright,
	XK_a, XK_s, XK_d, XK_f,
	XK_g, XK_h, XK_j, XK_k,
    XK_l, XK_semicolon, XK_apostrophe, XK_z,
	XK_x, XK_c, XK_v, XK_b,
	XK_n, XK_m, XK_comma, XK_period, 
	XK_slash,	
    //Special Character 2
    XK_space, XK_Tab
  };
  
  char sym[] = { 
      //Shift Numeric 14
      '~', '!', '@', '#', '$',
	  '%', '^', '&', '*', '(',
	  ')', '_', '+', '|',
	  //Numeric 14
	  '`', '1', '2', '3', '4',
	  '5', '6', '7', '8', '9',
	  '0', '-', '=', '\\',
	  //Shift Character 33
	  'Q', 'W', 'E', 'R',
	  'T', 'Y', 'U', 'I',
	  'O', 'P', '{', '}',
	  'A', 'S', 'D', 'F',
	  'G', 'H', 'J', 'K',
	  'L', ':', '\"', 'Z',
	  'X', 'C', 'V', 'B',
	  'N', 'M', '<', '>',
	  '?',
	  //Character 33
	  'q', 'w', 'e', 'r',
	  't', 'y', 'u', 'i',
	  'o', 'p', '[', ']',
	  'a', 's', 'd', 'f',
	  'g', 'h', 'j', 'k',
	  'l', ';', '\'', 'z',
	  'x', 'c', 'v', 'b',
	  'n', 'm', ',', '.',
	  '/', 
	  //Character 2
	  ' ', '	'
};
  
  for (iterator = 0; iterator < 96; iterator++)
    if (ch == sym[iterator]) {
      res = eng[iterator]; 
      break;
    }
  return res;
}  
  
/*KeySym xevent_char_to_keysym(char ch);
KeySym xevent_char_to_keysym(char ch)
{
  num iterator;
  KeySym res = None;
  KeySym ks[] = { XK_quotedbl, XK_apostrophe, XK_bracketleft, XK_bracketright,
		  XK_comma, XK_period, XK_colon, XK_semicolon,
		  XK_less, XK_greater, XK_braceleft, XK_braceright,
		  XK_space, XK_exclam, XK_at, XK_numbersign,
		  XK_ssharp, XK_dollar, XK_percent, XK_degree,
		  XK_ampersand, XK_asterisk, XK_parenleft, XK_parenright,
		  XK_underscore, XK_equal, XK_plus, XK_minus,
		  XK_slash, XK_bar, XK_backslash, XK_question
  };
  char sym[] = { '"', '\'', '{', '}',
		',', '.', ':', ';',
		'<', '>', '[', ']',
		' ', '!', '@', '#',	//TODO
		'#', '$', '%', '^',	//TODO
		'&', '*', '(', ')',
		'_', '=', '+', '-',
		'/', '|', '\\', '?'
  };

  for (iterator = 0; iterator < 32; iterator++)
    if (ch == sym[iterator]) {
      res = ks[iterator];
      break;
    }
  return res;
}*/

int xevent_uninit(struct _xevent *p);
int xevent_uninit(struct _xevent *p)
{
  status err = SUCCESS;
  
  if (p) {
    if ((p->communicator->uninit(p->communicator)) != SUCCESS);
    err = WARNING;
    free (p);
  }
  else
    err = WARNING;
  
  return err;
}

xevent* xevent_init(xevent *p)
{
  p = NULL;
  
  p = (xevent *) xnmalloc(sizeof(struct _xevent));
  if (!p)
    return NULL;
  p->communicator = xobj_init (p->communicator);
  if (NULL == p->communicator) {
    p->communicator->uninit(p->communicator);
    free (p);
    return NULL;
  }
  
  /*functions mapping*/
  p->set_communicator = xevent_set_communicator;
  p->get_communicator_win = xevent_get_communicator_win;
  p->get_event_win = xevent_get_event_win;
  p->update_by_focus = xevent_update_by_focus;
  p->send_xkey = xevent_send_xkey;
  p->send_event = xevent_send_event;
  p->send_string = xevent_send_string;
  p->put_back_to_queue = xevent_put_back_to_queue;
  p->chck_key_release = xevent_chck_key_release;
  p->chck_key_press = xevent_chck_key_press;
  p->chck_destroy_notify = xevent_chck_destroy_notify;
  p->is_native = xevent_is_native;
  p->get_curr_char = xevent_get_curr_char;
  p->get_curr_keysym = xevent_get_curr_keysym;
  p->get_curr_modifier = xevent_get_curr_modifier;
  //p->ruchar_to_engchar = xevent_ruchar_to_engchar;
  //p->RuWinlayout_ks_to_ENlayout_ks = xevent_RuWinlayout_ks_to_ENlayout_ks;
  p->char_to_keysym = xevent_char_to_keysym;
  p->keysym_to_char = xevent_keysym_to_char;
    p->uninit = xevent_uninit;

  return p;
}

/*--------xselection--------*/

int xselection_request_to_convert(struct _xselection *p);
int xselection_request_to_convert(struct _xselection *p)
{
  Window Sown;
  XEvent e;
  unsigned long len, bytes_left, dummy;
  unsigned char *data;
  int format, result;
  Atom type;
  
  //unsure delete recieve property
  XDeleteProperty (p->requestor->xDisplay,
		   p->requestor->xWindow, p->property);
  //send convert selection request
  p->target = XA_UTF8_STRING (p->requestor->xDisplay);
  Sown = XGetSelectionOwner (p->requestor->xDisplay, XA_PRIMARY);
  printf ("Selection owner%i\n", (int)Sown);
  XConvertSelection (p->requestor->xDisplay, XA_PRIMARY, XA_STRING, None,
			p->requestor->xWindow, CurrentTime);  // w == our window's ID
  XFlush (p->requestor->xDisplay);
  XNextEvent (p->requestor->xDisplay, &e);
  if (e.type == SelectionNotify) {
	printf ("Notified about XConvertSelection With Property %d\n", (int) e.xselection.property);
	if ((e.xselection.property == XA_STRING)||(e.xselection.property == 81)) {
	  printf ("It is not 'None' so it is Ok\n");
	  XGetWindowProperty (p->requestor->xDisplay, p->requestor->xWindow, e.xselection.property, 0, 0, 0, AnyPropertyType,
					&type, &format,	&len, &bytes_left, &data);		
	  printf ("type:%i len:%i format:%i byte_left:%i\n", 
				(int)type, len, format, bytes_left);
	  // DATA is There
	  if (bytes_left > 0)
	  {
	  	result = XGetWindowProperty (p->requestor->xDisplay, p->requestor->xWindow, e.xselection.property, 0,bytes_left,0,
			AnyPropertyType, &type,&format,	&len, &dummy, &data);
      	if (result == Success)
		  printf ("DATA IS HERE!!```%s'''\n", data);
		else printf ("FAIL\n");
		  if (data != NULL)	XFree (data);
	  }
	  
	}	
  }
  else
	return ERROR;  
  /*XConvertSelection (p->requestor->xDisplay,
		     p->selection, p->target, p->property,
		     p->requestor->xWindow, p->currentTime);*/

  
}

XSelectionEvent* xselection_wait_for_notify(struct _xselection *p);
XSelectionEvent* xselection_wait_for_notify(struct _xselection *p)
{
  XEvent event;
  XSelectionEvent *sevent = (XSelectionEvent *) malloc (sizeof (XSelectionEvent));
  while ((event.type != SelectionNotify)
	 && (event.xselection.requestor != xobj_get_win (p->requestor)))
    XNextEvent (p->requestor->xDisplay, &event);
  if (event.xselection.property != None) {
    *sevent = event.xselection;
    return sevent;
  }
  else
    send_error (DEBUG, "Convert to selection target return None answer\n", NULL);

  return NULL;
}

char* xselection_get_to_string(struct _xselection *p,XSelectionEvent *xselt);
char* xselection_get_to_string(struct _xselection *p,XSelectionEvent *xselt)
{
  int format;
  unsigned long len, bytes_left = 1, offset, part_len;
  unsigned char *data;
  int result;
  char *string = NULL;
  Atom type;
  Display *ManagedDisplay = p->owner->get_display (p->owner);
  Window ManagedWindow = xobj_get_win(p->owner);

  part_len = 8;
  offset = 0;
  result = XGetWindowProperty (ManagedDisplay, ManagedWindow,
			       xselt->property,
			       offset, part_len,
			       FALSE, AnyPropertyType,
			       &type, &format, &len, &bytes_left, &data);
  if (result == Success) {
    string = (char *) malloc ((format / 8) * (len + bytes_left) + 1);
    string = memset (string, 0, (format / 8) * (len + bytes_left) + 1);
    string = strcat (string, (char*)data);
    offset += part_len;
  }
  else
    return NULL;
  while (bytes_left > 0) {
    //try to get minimum portion of data from our recieve property to get the size of data (len*(format/8) )
    result = XGetWindowProperty (ManagedDisplay, ManagedWindow,
				 xselt->property,
				 offset, part_len,
				 FALSE, AnyPropertyType,
				 &type, &format, &len, &bytes_left, &data);

    if (result != Success) {
      send_error (ERROR, "Fail to get selected text\n", NULL);
      if (string)
	    free (string);
      return NULL;
    }
    else {
      string = strcat (string, (char *) data);
      offset += part_len;
    }
    XFree (data);
  }
  //We've already finish working with our recieve property. Notify everybody we'll done.
  XDeleteProperty (ManagedDisplay, ManagedWindow, xselt->property);

  return string;
}
 
unsigned char* xselection_get(struct _xselection *p);
unsigned char* xselection_get(struct _xselection *p)
{
  char *string;
  Window Sown;
  XEvent e, respond;
  unsigned long len, bytes_left, dummy;
  unsigned char *data = NULL;
  int format, result;
  Atom type;

  //send convert selection request
  Atom UTF8_STRING_Atom, NONE_Atom;
  UTF8_STRING_Atom = XInternAtom(p->requestor->xDisplay, "UTF8_STRING",FALSE);
  NONE_Atom = XInternAtom(p->requestor->xDisplay, "NONE",FALSE);

  p->target = XA_UTF8_STRING (p->requestor->xDisplay);

  Sown = XGetSelectionOwner (p->requestor->xDisplay, XA_PRIMARY);
  XConvertSelection (p->requestor->xDisplay, XA_PRIMARY, p->target, None,
			p->requestor->xWindow, CurrentTime);  // w == our window's ID
  XFlush (p->requestor->xDisplay);
  XNextEvent (p->requestor->xDisplay, &e);
  if (e.type == SelectionNotify) {
	if ((e.xselection.property == UTF8_STRING_Atom)||(e.xselection.property == NONE_Atom)) {
	  XGetWindowProperty (p->requestor->xDisplay, p->requestor->xWindow, e.xselection.property, 0, 0, 0, AnyPropertyType,
					&type, &format,	&len, &bytes_left, &data);		
	  // DATA is There
	  if (bytes_left > 0)
	  {  
	  	result = XGetWindowProperty (p->requestor->xDisplay, p->requestor->xWindow, e.xselection.property, 0,bytes_left,0,
			AnyPropertyType, &type,&format,	&len, &dummy, &data);
      	if (result == Success) {
		  return data;
		}	
		else {
		  send_error(DEBUG, "Get Window Select Property failed!\n", NULL);
		  return NULL;
		}	
	  }
	  else {
		send_error(DEBUG, "Get Window Select Len Text failed!\n", NULL);  
		return NULL;	
	  }	  
	}
    else {
	  send_error(DEBUG, "Get Window Property Type failed!\n", NULL);	
	  return NULL;	
	}	
  }
  else {
	send_error(DEBUG, "Select Notyfication  failed!\n", NULL);  
	return NULL; 
  }	  	  
}

int xselection_set_to_none(struct _xselection *p);
int xselection_set_to_none(struct _xselection *p)
{
  XSetSelectionOwner (p->owner->get_display (p->owner), p->selection, None, CurrentTime);
  return SUCCESS;
}

int xselection_wait_owner(struct _xselection *p,int time);
int xselection_wait_owner(struct _xselection *p,int time)
{
  num timeout = 0;
  while (++timeout != time / MIN_CYCLE_TIME) {
    if (XGetSelectionOwner
	(p->owner->get_display (p->owner), p->selection) != None)
      break;
    else
      usleep (MIN_CYCLE_TIME);
  }
  if (timeout == time / MIN_CYCLE_TIME)
    return ERROR;
  else
    return SUCCESS;
}

int xselection_uninit(struct _xselection *p);
int xselection_uninit(struct _xselection *p)
{
  status err = SUCCESS;

  if (p) {
    if (((p->owner->uninit(p->owner)) != SUCCESS)
	|| ((p->requestor->uninit(p->requestor)) != SUCCESS))
      err = WARNING;
    free (p);
  }
  else
    err = WARNING;
  return err;
}

xselection* xselection_init(xselection *p,Display *ManagedDisplay,Window ManagedWindow)
{
  p = NULL;

  p = (xselection *) xnmalloc(sizeof(struct _xselection));
  if (!p)
    return NULL;
  p->owner = xobj_init (p->owner);
  p->owner->xDisplay = ManagedDisplay;
  p->owner->xWindow = ManagedWindow;
  p->requestor = xobj_init(p->requestor);
  p->requestor->xDisplay = ManagedDisplay;
  p->requestor->xWindow = ManagedWindow;
  if (!(p->owner) || !(p->requestor)) {
    xobj_uninit (p->owner);
    xobj_uninit (p->requestor);
    free (p);
    return NULL;
  }

  p->selection = XA_PRIMARY;
  p->property = XInternAtom (ManagedDisplay, "_selected_STRING", FALSE);

  if (!X_HAVE_UTF8_STRING) {
    send_error (ERROR, "XServer doesn't support UTF8\n", NULL);
    return NULL;
  }

  /*functions mapping*/
  p->request_to_convert = xselection_request_to_convert;
  p->wait_for_notify = xselection_wait_for_notify;
  p->get_to_string = xselection_get_to_string;
  p->get = xselection_get;
  p->set_to_none = xselection_set_to_none;
  p->wait_owner = xselection_wait_owner;
  p->uninit = xselection_uninit;

  return p;
}

/*--------xswitchlang--------*/

/*TODO: working with many languages.*/
int xswitchlang_get_curr_lang(struct _xswitchlang *p, Display *dpy);
int xswitchlang_get_curr_lang(struct _xswitchlang *p, Display *dpy)
{
  int xkb_group;
  xkb_group = p->getActiveKeyboard(p);
  if (xkb_group == p->rus_win_keyboard_group) {
	return RUSSIAN;
  }
  else if (xkb_group == p->usa_eng_keyboard_group) {
	return ENGLISH;
  }
  else
	return -1;  
}

// Work with Keyboard Layout
char** xswitchlang_getKeyboardList (int* length, struct _xswitchlang *p);
char** xswitchlang_getKeyboardList (int* length, struct _xswitchlang *p) {
	
	int numGroups = 0;	
	XkbDescPtr xkb_desc = XkbGetMap (p->xobject->xDisplay, XkbKeySymsMask, XkbUseCoreKbd);
	XkbGetNames (p->xobject->xDisplay, XkbGroupNamesMask, xkb_desc);
	send_error(DEBUG,"Keyboard Layout Present In System:\n",NULL);

	while (xkb_desc->names->groups[numGroups]) numGroups++; //should be easy
	if (numGroups > 0) {
		char** groups = (char**) xnmalloc (numGroups);
		int i;
		for (i = 0; i < numGroups; i++) {
			char* name = XGetAtomName (p->xobject->xDisplay, xkb_desc->names->groups[i]);
            send_error(DEBUG,"  '%s' ",&name);
	        if (i == p->usa_eng_keyboard_group) {
	          printf("- must be ENGLISH (group %d) (if NOT, see configuration file)!\n", i);
	        }
	        else if (i == p->rus_win_keyboard_group) {
	          printf("- must be RUSSIAN (group %d) (if NOT, see configuration file)!\n", i);
	        }
	        else
	          printf("- (group %d)!\n");
		}
		*length = numGroups;
		
		return groups;
	}
	*length = 0;
	return NULL;
}

/* 
* Sets the active keyboard layout.
* Can be used with getKeyboardList() to implement BidiUtil.setKeyboardLanguage().
*/
void xswitchlang_setActiveKeyboard (int group, struct _xswitchlang *p);
void xswitchlang_setActiveKeyboard (int group, struct _xswitchlang *p) {
	XkbLockGroup (p->xobject->xDisplay, XkbUseCoreKbd, group);
}

/* 
* Returns the group number of the active keyboard 
* Can be used to implement Bidi.getKeyboardLanguage()
*/
int xswitchlang_getActiveKeyboard (struct _xswitchlang *p);
int xswitchlang_getActiveKeyboard (struct _xswitchlang *p) {
	XkbStateRec xkbState;
	XkbGetState(p->xobject->xDisplay, XkbUseCoreKbd, &xkbState);
	return xkbState.group;
}

int xswitchlang_switch_lang(xswitchlang *p);
int xswitchlang_switch_lang(xswitchlang *p)
{
  status err = SUCCESS;
  int xkb_group;
  xkb_group = p->getActiveKeyboard(p);	

  if (xkb_group == p->rus_win_keyboard_group) {
	p->setActiveKeyboard(p->usa_eng_keyboard_group, p);
    p->currlang = ENGLISH;
  }	  
  else if (xkb_group == p->usa_eng_keyboard_group){
    p->setActiveKeyboard(p->rus_win_keyboard_group, p);
	p->currlang = RUSSIAN;	  
  }
  else
    send_error (WARNING, "Unknown xkb group\n", NULL);
  
  return err;
}

int xswitchlang_uninit(struct _xswitchlang *p);
int xswitchlang_uninit(struct _xswitchlang *p)
{
  status err = SUCCESS;

  if (!p)
    err = NULL_PTR;
  else {
    if (p->xobject) {
	  p->xobject->uninit(p->xobject);	
	}	
    else
      err = NULL_PTR;
    free (p);
  }

  return err;
}

xswitchlang *xswitchlang_init(struct _xswitchlang *p,Display * xDisplay);
xswitchlang *xswitchlang_init(struct _xswitchlang *p,Display * xDisplay)
{
  status err;
  p = NULL;

  p = (xswitchlang *)xnmalloc(sizeof(struct _xswitchlang));
  if (!p)
    return NULL;

  p->xobject = xobj_init(p->xobject);
  if (!(p->xobject)) {
    free (p);
    return NULL;
  }
  p->xobject->xDisplay = xDisplay;
  if (ERROR == p->xobject->connect (p->xobject)) {
	send_error (ERROR, "Connect to XServer Failed\n", NULL);  
    return NULL;
  }
  err = XkbQueryExtension (xDisplay, NULL, NULL, NULL, NULL, NULL);
  if (!err)
    return NULL;

  //XkbGetState (xDisplay, XkbUseCoreKbd, &(p->state));

  /*functions mapping*/
  p->switch_lang = xswitchlang_switch_lang;
  p->get_curr_lang = xswitchlang_get_curr_lang;
  p->getKeyboardList = xswitchlang_getKeyboardList;
  p->setActiveKeyboard = xswitchlang_setActiveKeyboard;
  p->getActiveKeyboard = xswitchlang_getActiveKeyboard;
    

  p->uninit = xswitchlang_uninit;

  return p;
}

/*--------xbindingtable--------*/

int xbindingtable_set_action(struct _xbindingtable *p,int action);
int xbindingtable_set_action(struct _xbindingtable *p,int action)
{
  if ((action < 0) || (action > p->ibind - 1))
    return INCORRECT_OPERATION;

  p->current_action = action;
  return SUCCESS;
}

int xbindingtable_get_action(struct _xbindingtable *p);
int xbindingtable_get_action(struct _xbindingtable *p)
{
  return p->current_action;
}

int xbindingtable_bind_key(struct _xbindingtable *p,KeySym key,KeySym alt,KeySym modifier,int action);
int xbindingtable_bind_key(struct _xbindingtable *p,KeySym key,KeySym alt,
		 KeySym modifier,int action)
{
  if ((action < 0) || (action > p->ibind - 1))
    return INCORRECT_OPERATION;

  p->bindlist[action].key = key;
  p->bindlist[action].key_alt = alt;
  p->bindlist[action].modifier = modifier;

  return SUCCESS;
}

int xbindingtable_find_bind_action(struct _xbindingtable *p,KeySym key,KeySym modifier);
int xbindingtable_find_bind_action(struct _xbindingtable *p,KeySym key,
			KeySym modifier)
{
  num itr;
  status action = -1;

  for (itr = 0; itr < p->ibind; itr++) {
    if (((p->bindlist[itr].key == key)
	 && (p->bindlist[itr].modifier == modifier))
	|| ((p->bindlist[itr].key_alt == key)
	    && (p->bindlist[itr].modifier == modifier))) {
      action = itr;	  
      break;
    }
	
  }	
  return action;
}

int xbindingtable_uninit(struct _xbindingtable *p);
int xbindingtable_uninit(struct _xbindingtable *p)
{
  status err = SUCCESS;

  if (p!=NULL) {
    if (p->bindlist != NULL)
      free (p->bindlist);
    else
      err = WARNING;
    free (p);
  }
  else
    err = WARNING;

  return SUCCESS;
}

xbindingtable* xbindingtable_init(xbindingtable *p,int ibind)
{
  p = NULL;

  if (ibind < 1)
    return NULL;
  p = (xbindingtable *)xnmalloc(sizeof(struct _xbindingtable));
  if (p == NULL)
    return NULL;
  p->ibind = ibind;
  p->bindlist = NULL;
  p->bindlist = (xbinding *) xnmalloc(sizeof(struct _xbinding));
  if (p->bindlist == NULL) {
    free (p);
    return NULL;
  }
  p->bindlist = memset (p->bindlist, 0, ibind * sizeof (struct _xbinding));
  p->current_action = -1;

  /*functions mapping*/
  p->set_action = xbindingtable_set_action;
  p->get_action = xbindingtable_get_action;
  p->bind_key = xbindingtable_bind_key;
  p->find_bind_action = xbindingtable_find_bind_action;
  p->uninit = xbindingtable_uninit;

  return p;
}
